#!/usr/bin/python2
# -*- coding: utf-8 -*-

from snapconnect import snap
from time import time
from time import sleep
import datetime
import logging
import ttyusb
import sys
import random
import string
import re
from binascii import hexlify
from binascii import unhexlify
from snaplib import ScriptsManager
from snaplib import SnappyUploader
import os
import collections
try:
    from _version import __version__, __revision__
except ImportError:
    __version__ = "UNKNOWN"
    __revision__ = "UNKNOWN"


SPY_UPLOAD_TIMEOUT = -100
PART_NUMBERS = {
    "SM220UF1": ";SM220UF1;",
    "SM220UF1-DMX": ";SM220UF1-DMX;",
    "SM220UF1-XSBT": ";SM220UF1-XSBT;",
    "RF200PF1": "^001C2C1F00",
    "RF200PD1": "^001C2C1AD6",
    "RF220UF1": "^001C2C1B26",
    "RF220SU": "^001C2C2AD6",
    "RF220SU-EU": "^001C2C2ED6",
    }


def barcode2partnumber(barcode):
    """
    SM220UF1
        14491001;260098C7;SM220UF1;0202
    SM220UF1-DMX
        14491001;260098FD;SM220UF1-DMX;0202
    SM220UF1-XSBT
        14491001;260098FD;SM220UF1-XSBT;0202
    RF200PF1
        001C2C1F0005F6A9
    RF200PD1
        001C2C1AD606C4FB
    RF220UF1
        001C2C1B26066BC4
    RF220SU
        001C2C2AD6111111
    RF220SU-EU
        001C2C2ED6123ABC
    """
    for key, value in PART_NUMBERS.iteritems():
        if re.search(value, barcode):
            logging.info(
                "barcode %s is recognized as part number %s",
                repr(barcode),
                repr(key)
                )
            return key


def randomword(length):
    return ''.join(random.choice(string.lowercase) for i in range(length))


class UpgradeFailedException(Exception):
    pass


class SnapEngine(snap.Snap):
    def __init__(
            self,
            port=None,
            expected_address=None,
            timeout=3,
            ports=None,
            addr='\x00\x00\x20',
            license_file='License.dat',
            nvparams_file='nvparams.dat'):
        super(SnapEngine, self).__init__(
            funcs={},
            license_file=license_file,
            nvparams_file=nvparams_file,
            addr=addr
            )
        self.opened_serial_ports = []
        self.open_ports = {}
        self.retry = 15
        self.timeout = timeout
        # self.add_rpc_func('callback_function', self.callback_function)
        if port:
            logging.info('opening: %r', port)
            self.open_serial(snap.SERIAL_TYPE_RS232, port)
        elif ports:
            for port in ports:
                logging.info('opening: %r', port)
                self.open_serial(snap.SERIAL_TYPE_RS232, port)
        if expected_address:
            self.local_address = self.call(None, 'localAddr')
            if expected_address != self.local_address:
                raise Exception(
                    'ERROR: expected address is %s, but found %s on port %s',
                    (expected_address, self.local_address, port))
        # True WHILE a file upload is in progress
        self.progInProgress = False
        # Time to wait for a remote device to respond to upload requests
        self.wait_time = 100
        # Timer used to tell if commands are taking too long
        # (during the programming process)
        self.respTimer = 0
        # Flag that keeps the timer tick from rescheduling itself
        self.stopTimer = False
        # The following get filled in on-the-fly
        # Filled in by beginUpload()
        # The file to be transferred
        self.fileName = ''
        # SNAP Address of the remote device to update
        self.targetAddr = None
        self.errCode = None
        self.add_rpc_func(
            "tellVmStat",
            self.tellVmStat)
        self.add_rpc_func(
            "su_recvd_reboot",
            self.su_recvd_reboot)
        self.set_hook(
            snap.hooks.HOOK_OTA_UPGRADE_COMPLETE,
            self.upgrade_complete_hook)
        self.set_hook(
            snap.hooks.HOOK_OTA_UPGRADE_STATUS,
            self.upgrade_status_hook)
        self.status_time = None
        self.response_received = None
        self.upgrade_status = None
        self.request_sent = None
        self.response_time = None
        self.function_destination = None
        self.filename = None
        self.return_value = None
        self.number_of_requests = 0

    def upgrade_complete_hook(self, addr, status, message):
        self.upgrade_status = status
        if status == snap.OTA_PROGRESS_COMPLETE:
            logging.info(
                "Upgrade is complete for node:%r",
                hexlify(addr),
                )
        else:
            logging.error(
                "Upgrade failed for %r: %r (Status=%r)",
                addr,
                message,
                status,
                )
            raise UpgradeFailedException()

    def upgrade_status_hook(self, addr, percent_complete):
        # This displays a percentage complete,
        # but it prints a lot so you might not want it on
        time_now = time()
        if (self.status_time is None) or ((time_now - self.status_time) > 1):
            self.status_time = time_now
            logging.info(
                "Upgrade progressed for node:%r %.1f",
                hexlify(addr),
                percent_complete,
                )

    def firmware_upgrade(
            self, mac, firmware_file, unreset_delay=None, unreset_func=None):
        self.status_time = None
        self.upgrade_status = None
        val = None
        upgrade = self.upgrade_firmware(mac, firmware_file)
        unreset_time = time() + unreset_delay if unreset_delay else None
        if not upgrade:
            raise Exception("Something went wrong while starting the upgrade!")
        while self.upgrade_status is None:
            self.poll()
            if unreset_time:
                if time() >= unreset_time:
                    self.info('calling unreset')
                    val = unreset_func() if isinstance(
                        unreset_func, collections.Callable
                        ) else unreset_func[0](*unreset_func[1:])
                    unreset_time = None
        return val

    def tellVmStat(self, arg, val):
        try:
            # print '.',
            self.spy_upload_mgr.onTellVmStat(
                self.rpc_source_addr(), arg, val)
        except AttributeError:
            pass

    def su_recvd_reboot(self, dummy):
        try:
            self.spy_upload_mgr.on_recvd_reboot(
                self.rpc_source_addr())
        except AttributeError:
            pass

    def upload_spy(
            self, mac, spy_file, unreset_delay=None, unreset_func=None,
            wait_time=100):
        self.call_log('spy', hexlify(mac), spy_file)
        self.wait_time = wait_time
        self.errCode = None
        val = None
        self.beginUpload(mac, spy_file)
        unreset_time = time() + unreset_delay if unreset_delay else None
        while self.progInProgress:
            self.poll()
            if unreset_time:
                if time() >= unreset_time:
                    self.info('calling unreset')
                    val = unreset_func() if isinstance(
                        unreset_func, collections.Callable
                        ) else unreset_func[0](*unreset_func[1:])
                    unreset_time = None
        if self.errCode != snap.SNAPPY_PROGRESS_COMPLETE:
            raise Exception(
                "spy_uploader_errCode = %s" %
                repr(self.errCode))
        return val

    def beginUpload(self, destAddr, filename):
        """
        Function that kicks off the script upload process
        Requires a SNAP Connect instance to be in place
        Requires a callback instance to be in place
        Parameters:
            destAddr  - the SNAP Address of the SNAP Node
                        to be uploaded with the script (SPY file)
            filename  - the exact file name (including path if needed)
                        of the SPY file to be uploaded
        """
        self.targetAddr = destAddr
        self.filename = filename
        self._startupload()

    def _startupload(self):
        """
        Called internally for every upload attempt.
        You should be calling beginUpload()
        """
        f = None
        uploadImage = None
        if not os.path.isfile(self.filename):
            raise ValueError('%s is not a file' % repr(self.filename))
        with open(self.filename, 'rb') as f:
            file_data = f.read()
        uploadImage = ScriptsManager.getSnappyStringFromExport(file_data)
        # Kickoff upload
        if uploadImage:
            self.respTimer = self.wait_time
            self.scheduler.schedule(1.0, self.timerTick)

            upload = self.spy_upload_mgr.startUpload(
                self.targetAddr, uploadImage)
            upload.registerFinishedCallback(self.app_upload_hook)
            self.progInProgress = True

    def app_upload_hook(self, snappy_upload_obj, result):
        if result == SnappyUploader.SNAPPY_PROGRESS_COMPLETE:
            logging.info('SPY upload success')
        elif result == snap.SNAPPY_PROGRESS_TIMEOUT:
            logging.error('SPY upload timeout error')
        else:
            logging.error('SPY upload error')
        self.completeProcess(result)

    def completeProcess(self, errCode):
        """
        Process has completed (either in failure or success)
        Reset for the next time
        """
        # Keep the next timer tick from rescheduling itself
        self.stopTimer = True
        self.progInProgress = False
        self.respTimer = 0  # Timer to tell if commands are taking too long
        self.targetAddr = None  # SNAP Address of the remote device to update
        self.fileName = ''  # Filename too
        self.errCode = errCode

    def timerTick(self):
        """
        Function called by timing mechanism on a regular basis
        Keeps track of when retries are needed
        """
        if self.progInProgress is True:
            self.respTimer -= 1
            logging.info("timerTick %d", self.respTimer)
            if self.respTimer == 0:
                # Timeout uploading
                logging.error(
                    "Spy upload timeout to node: %s",
                    hexlify(self.targetAddr)
                    )
                self.completeProcess(SPY_UPLOAD_TIMEOUT)
                return False  # Do not reschedule the tick
        if self.stopTimer:  # The way to stop the timer if necessary
            self.stopTimer = False
            return False
        return True  # Reschedule yourself

    def call_cb(self, cb, dest, function_name, *args):
        cb_name = randomword(10)
        self.add_rpc_func(cb, self.callback_function)
        try:
            return self.call1(cb_name, dest, function_name, *args)
        finally:
            del self.local_dispatcher.funcs[cb]

    def call(self, dest, function_name, *args):
        cb_name = randomword(10)
        self.add_rpc_func(cb_name, self.callback_function)
        try:
            return self.call1(cb_name, dest, function_name, *args)
        finally:
            del self.local_dispatcher.funcs[cb_name]

    def call1(self, cb_name, dest, function_name, *args):
        # sys.stderr.write('call %s%s\n' % (function_name, repr(args)))
        # sys.stderr.write(str(self) + '\n')
        self.response_received = None
        self.function_destination = dest
        self.number_of_requests = 0
        while True:
            if self.number_of_requests > self.retry:
                raise Exception('ERROR: timeout calling %s' % function_name)
            self.number_of_requests += 1
            if self.number_of_requests > 1:
                # print 'request #%d calling %s' % (
                    # self.number_of_requests, function_name)
                if self.function_destination:
                    sys.stderr.write(
                        'request #%d: %s is calling: rpc(%s, %s, ...)\n' %
                        (
                            self.number_of_requests,
                            hexlify(self.local_addr()),
                            hexlify(self.function_destination),
                            function_name
                        )
                    )
            if self.function_destination:
                # sys.stderr.write(
                    # '%s calling: rpc %s\n' %
                    # (hexlify(self.local_addr()),
                        # hexlify(self.function_destination)))
                self.call_log(
                    'rpc',
                    hexlify(self.function_destination),
                    function_name,
                    *args)
                self.rpc(
                    self.function_destination,
                    'callback',
                    cb_name,
                    function_name, *args)
            else:
                # sys.stderr.write('mcast_rpc...\n')
                self.call_log(
                    'mcast_rpc', None, function_name, *args)
                self.mcast_rpc(
                    1, 1, 'callback', cb_name,
                    function_name, *args)
            self.request_sent = time()
            function_response_timeout = self.timeout + self.request_sent
            while True:
                if time() >= function_response_timeout:
                    break
                self.poll()
                if self.response_received:
                    self.response_time = \
                        self.response_received - self.request_sent
                    # print 'response_time =', self.response_time
                    return self.return_value

    def call_log(self, call_type, target_address, function_name, *args):
        call_log_file_name = os.path.expanduser('~/.call_log.csv')
        if os.path.isfile(call_log_file_name):
            line = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")
            line += ',' + repr(call_type)
            line += ',' + repr(target_address)
            line += ',' + repr(function_name)
            for arg in args:
                line += ',' + repr(arg)
            with open(call_log_file_name, "ab") as call_log_file:
                call_log_file.write(line + '\n')

    def get_info(self, macinfo, my_id=None):
        info = {}
        if my_id is not None:
            info['id'] = my_id
        info['name'] = self.call(macinfo, 'loadNvParam', 8)
        info['mac'] = hexlify(self.call(macinfo, 'loadNvParam', 2))
        info['channel'] = self.call(macinfo, 'loadNvParam', 4)
        nid = self.call(macinfo, 'loadNvParam', 3)
        nid_str = ('0x%.2x%.2x' % ((nid >> 8) % 256, nid % 256)) if nid else ''
        info['networkId'] = nid_str
        fea = self.call(macinfo, 'loadNvParam', 11)
        fea_str = ('0x%.2x%.2x' % ((fea >> 8) % 256, fea % 256)) if fea else ''
        info['feature'] = fea_str
        return info

    def callback_function(self, return_value):
        # sys.stderr.write('callback_function return_value: %s\n' %
        # repr(return_value))
        # sys.stderr.write(str(self) + '\n')
        # sys.stderr.write('local_addr: %s\n' %
            # hexlify(self.local_addr()))
        if (
                self.function_destination is None or
                self.rpc_source_addr() == self.function_destination):
            self.response_received = time()
            self.return_value = return_value

    def open_serial(
            self,
            serial_type=snap.SERIAL_TYPE_RS232,
            port=None,
            vid=None,
            pid=None,
            serial_number=None,
            reconnect=False,
            dll_path=None,
            forward_groups=None,
            baudrate=None,
            io_loop=None,
            close_all=True):
        if port is None:
            if vid is None:
                raise ValueError("vid cannot be None if port is None")
            if pid is None:
                raise ValueError("pid cannot be None if port is None")
            port = ttyusb.find_tty_usb_sn(vid, pid, serial_number)
        if close_all:
            self.close_all_serial()
        if dll_path is not None:
            super(SnapEngine, self).open_serial(
                serial_type,
                port,
                reconnect=reconnect,
                dll_path=dll_path,
                forward_groups=forward_groups,
                baudrate=baudrate,
                io_loop=io_loop
                )
        else:
            super(SnapEngine, self).open_serial(
                serial_type,
                port,
                reconnect=reconnect,
                forward_groups=forward_groups,
                baudrate=baudrate,
                io_loop=io_loop
                )
        self.opened_serial_ports.append(port)
        return port

    def close_all_serial(self):
        self.open_ports = {}
        for port in self.opened_serial_ports:
            self.opened_serial_ports.remove(port)
            try:
                self.close_serial(snap.SERIAL_TYPE_RS232, port)
            except:
                pass

    def open_usb_serial(
            self,
            usb_serial,
            reconnect=False,
            dll_path=None,
            forward_groups=None,
            baudrate=None,
            io_loop=None,
            close_all=True):
        self.open_ports[usb_serial['name']] = self.open_serial(
            vid=usb_serial['vid'],
            pid=usb_serial['pid'],
            serial_number=usb_serial['sn'],
            reconnect=reconnect,
            dll_path=dll_path,
            forward_groups=forward_groups,
            baudrate=baudrate,
            io_loop=io_loop,
            close_all=close_all,
        )

    def close_usb_serial(self, usb_serial):
        port = self.open_ports.pop(usb_serial['name'], None)
        if port is not None:
            if port in self.opened_serial_ports:
                self.opened_serial_ports.remove(port)
            try:
                self.close_serial(snap.SERIAL_TYPE_RS232, port)
            except:
                pass

    def read_snappy_os_version(self, mac):
        version_major = self.call(mac, 'getInfo', 5)
        version_minor = self.call(mac, 'getInfo', 6)
        version_build = self.call(mac, 'getInfo', 7)
        return (version_major, version_minor, version_build)

    def read_snappy_os_version_string(self, mac):
        version_major = self.call(mac, 'getInfo', 5)
        version_minor = self.call(mac, 'getInfo', 6)
        version_build = self.call(mac, 'getInfo', 7)
        return (
            str(version_major) +
            '.' +
            str(version_minor) +
            '.' +
            str(version_build)
            )

    def enable_crc(self, mac, try_enabled_first=False, enable=True):
        mac_str = 'Unknown' if mac is None else hexlify(mac)
        retry_save = self.retry
        self.retry = 1
        comm_retry = 0
        while True:
            comm_retry += 1
            if comm_retry > retry_save:
                self.retry = retry_save
                raise Exception(
                    'comm failed with node: %s' % mac_str
                    )
            if try_enabled_first:
                try:
                    self.save_nv_param(11, 0x100)
                    nv11 = self.call(mac, 'loadNvParam', 11)
                    if enable:
                        break
                    self.call(
                        mac,
                        'saveNvParam',
                        11,
                        0xfeff & nv11
                        )
                    self.call(
                        mac,
                        'reboot',
                        )
                    sleep(2)
                    self.save_nv_param(11, 0)
                    break
                except:
                    logging.info(
                        'with CRC enabled, comm failed with node %s',
                        mac_str
                        )
                try:
                    self.save_nv_param(11, 0)
                    nv11 = self.call(mac, 'loadNvParam', 11)
                    if not enable:
                        break
                    self.call(
                        mac,
                        'saveNvParam',
                        11,
                        0x100 | nv11
                        )
                    self.call(
                        mac,
                        'reboot',
                        )
                    sleep(2)
                    self.save_nv_param(11, 0x100)
                    break
                except:
                    logging.info(
                        'with CRC disabled, comm failed with node %s',
                        mac_str
                        )
            else:
                try:
                    self.save_nv_param(11, 0)
                    nv11 = self.call(mac, 'loadNvParam', 11)
                    if not enable:
                        break
                    self.call(
                        mac,
                        'saveNvParam',
                        11,
                        0x100 | nv11
                        )
                    self.call(
                        mac,
                        'reboot',
                        )
                    sleep(2)
                    self.save_nv_param(11, 0x100)
                    break
                except:
                    logging.info(
                        'with CRC disabled, comm failed with node %s',
                        mac_str
                        )
                try:
                    self.save_nv_param(11, 0x100)
                    nv11 = self.call(mac, 'loadNvParam', 11)
                    if enable:
                        break
                    self.call(
                        mac,
                        'saveNvParam',
                        11,
                        0xfeff & nv11
                        )
                    self.call(
                        mac,
                        'reboot',
                        )
                    sleep(2)
                    self.save_nv_param(11, 0)
                    break
                except:
                    logging.info(
                        'with CRC enabled, comm failed with node %s',
                        mac_str
                        )
        self.retry = retry_save

    def rpcpoll(self, dest, function_name, *args):
        self.rpc(dest, function_name, *args)
        self.poll()

def firmware_upgrade(port, firmware_file, license_file, addr, version=None):
    start_time = time()
    # snapconnect_address = '\x00\x00' + chr(0x40 + int(x))
    # print 'snapconnect_address:', hexlify(snapconnect_address)
    eng = SnapEngine(license_file=license_file, addr=addr)
    eng.open_serial(snap.SERIAL_TYPE_RS232, port)
    eng.enable_crc(None, False, False)
    nv2_read = eng.call(None, 'loadNvParam', 2)
    node = nv2_read[-3:]
    node_nr = int(hexlify(node), 16)
    logging.info(
        'MAC: %s programming started...',
        hexlify(node)
        )
    eng.firmware_upgrade(node, firmware_file)
    if version is not None:
        sleep(0.5)
        version_read = eng.read_snappy_os_version_string()
    eng.close_all_serial()
    finish_time = time()
    elapsed = finish_time - start_time
    if version is not None:
        if version_read == version:
            logging.info(
                'MAC: %s elapsed: %s version_read: %s',
                hexlify(node),
                elapsed,
                version_read
                )
            return node_nr
        else:
            logging.error(
                'MAC: %s elapsed: %s version_read: %s version_expected: %s',
                hexlify(node),
                elapsed,
                version_read,
                version
                )
            return -node_nr
    logging.info(
        'MAC: %s elapsed: %s',
        hexlify(node),
        elapsed
        )
    return node_nr


def firmware_upgrade_command():
    logging.basicConfig(level=logging.INFO)
    if len(sys.argv) == 5:
        sys.exit(
            firmware_upgrade(
                sys.argv[1],
                sys.argv[2],
                sys.argv[3],
                unhexlify(sys.argv[4])
                )
            )
    elif len(sys.argv) == 6:
        sys.exit(
            firmware_upgrade(
                sys.argv[1],
                sys.argv[2],
                sys.argv[3],
                unhexlify(sys.argv[4]),
                sys.argv[5]
                )
            )
    else:
        logging.error('arguments are missing')
        sys.exit(-1)


def _main():
    logging.basicConfig(level=logging.INFO)
    logging.info('snapengine version: %s', __version__)
    logging.info('snapengine revision: %s', __revision__)


if __name__ == "__main__":
    _main()
